#!@@PERL@@
#
# Spong client monitoring script.  This runs various tests locally on a machine
# and sends those results to the Spong server machine where the information
# is summarized and displayed.  This script currently checks the following:
#
#    * disk  (check to make sure that disks are not filling up)
#    * cpu   (check to make sure the load is reasonable)
#    * procs (check to make sure certain key processes are running)
#    * msgs  (scan through logs looking for problems)
#
# History:
# (1) Ported bb-local.sh script to perl. (Ed Hill Feb 26, 1997)

use Sys::Hostname;
use Socket;

srand( time() ^ ($$ + ($$ << 15 )) );

if( $ARGV[0] eq "--debug" )   { $debug = 1;   shift; }
if( $ARGV[0] eq "--restart" ) { $restart = 1; shift; }
if( $ARGV[0] eq "--kill" )    { $kill = 1;    shift; }

$me        = "@@BINDIR@@/spong-client";
$conf_file = $ARGV[0] || "@@ETCDIR@@/spong.conf";
($HOST)    = gethostbyname(&Sys::Hostname::hostname());
$HOST      =~ tr/A-Z/a-z/;

# Set up some signal handlers to handle our death gracefully, and also listen
# for the HUP signal, and if we se that, we re-exec ourself.

$SIG{'QUIT'} = \&exit_handler;
$SIG{'HUP'}  = \&hup_handler;
$SIG{'PIPE'} = \&pipe_handler;
$SIG{'CHLD'} = \&pipe_handler;

# Load our configuration variables, including anything specific to the host
# that we are running on.

require $conf_file || die "Can't load $conf_file: $!";
if( -f "$conf_file.$HOST" ) {
   require "$conf_file.$HOST" || die "Can't load $conf_file.$HOST: $!";
} else {
   my $tmp = (split( /\./, $HOST ))[0];
   if( -f "$conf_file.$tmp" ) { # for lazy typist
      require "$conf_file.$tmp" || die "Can't load $conf_file.$tmp: $!";
   }
}
&debug( "configuration file(s) loaded" );

# If the user gives us the --restart or --kill flags, then we signal the
# currently running spong-client process, and tell it to either die, or
# re-exec itself (which causes it to re-read it's configuration files.

if( $restart || $kill ) {
   open( PID, "$SPONGTMP/spong-client.pid") || die "Can't find pid file: $!";
   my $pid = <PID>; chomp $pid;
   close PID;
   
   if( $restart ) { 
      &debug( "telling pid $pid to restart" ); kill( 'HUP', $pid ); }
   if( $kill ) { 
      &debug( "telling pid $pid to die" ); kill( 'QUIT', $pid );}

   exit(0);
}

# Write our pid to the spong tmp directory.

open( PID, ">$SPONGTMP/spong-client.pid" ) || 
   print STDERR "Can't create spong-client.pid file: $!\n";
print PID "$$\n"; 
close( PID );


# Pretty simple program, check the stuff that we are looking at, report it to
# the server, sleep, and do it all again...

while( 1 ) {
   &check_disk_space();
   &check_cpu_load();
   &check_processes();
   &check_log_files();

   &check_local() if defined &check_local;

   # Sleep about $SPONGSLEEP seconds - add a little bit of randomness so 
   # that things don't get in sync and pound the spong-server.

   my $sleep = int($SPONGSLEEP - (.05 * $SPONGSLEEP) + rand(.1 * $SPONGSLEEP));
   &debug( "sleeping for $sleep seconds" );
   sleep $sleep;
}

# Check disk space.

sub check_disk_space { 
   my $color    = "green";
   my( $summary, $message, $check, @problems, $large, $lpercent );

   open( PIPE, "$DF |" );

 DFPIPE:
   while( <PIPE> ) {
      if( m!^(\S+)\s.*?\s(\d+)\%\s+[^/]*(/.*)$! ) {
	 my( $rawfs, $percent, $name ) = ( $1, $2, $3 );

	 my $panic = $DFCRIT{$name} || $DFCRIT{$rawfs} || $DFCRIT{"ALL"};
	 my $warn  = $DFWARN{$name} || $DFWARN{$rawfs} || $DFWARN{"ALL"};

	 foreach $check ( @DFIGNORE ) { 
	    next DFPIPE if $rawfs =~ /$check/ || $name =~ /$check/;
	 }

	 if( $percent > $lpercent ) { $lpercent = $percent; $large = $name; }

	 $message .= $_;
	 if( $percent >= $panic ) { 
	    $color = "red"; 
	    push( @problems, "$name $percent" );
	 } elsif( $percent >= $warn )  { 
	    $color = "yellow" unless $color eq "red";
	    push( @problems, "$name $percent" );
	 }
      }
   }
   close( PIPE );

   # Collect the problems, and print a single message describing the problem(s)

   if( $#problems < 0 ) {
      $summary = "largest filesystem $large at $lpercent%";
   } elsif( $#problems == 0 ) {
      my( $disk, $percent) = (split( /\s+/, $problems[0] ));
      $summary = "$disk is $percent% full";
   } else {
      my( $gstr, $sstr );

      $summary = "multiple problems: ";
      foreach( @problems ) {
	 my( $disk, $percent) = (split( /\s+/, $_ ));
	 $summary .= "$disk ($percent%), "; 
      }
      chop $summary; chop $summary;
   }

   &debug( "disk - $color, $summary" );
   &status( $SPONGSERVER, $HOST, "disk", $color, $summary, $message );
}

# Check CPU load.

sub check_cpu_load { 
   my $color    = "green";
   my $uptime   = `$UPTIME`;
   my( $message, $procs );

   my $s = 'up\s+([^\,]+)\,.*\s(\d+) user.+?' . 
      '[0-9\.]+\,\s*([0-9\.]+)\,\s*([0-9\.]+)\s*$';
   my( $up, $users, $load ) = ( $uptime =~ /$s/ );

   open( PIPE, "$PS |" );
   while( <PIPE> ) { if( $procs++ < 11 ) { $message .= $_; } }
   close( PIPE );

   # Check to see if the system has only been up a short time, and if it has
   # only been less then an hour, send a page

   if( $up =~ /min/ )      { $color = "yellow"; }
   if( $load > $CPUWARN )  { $color = "yellow"; }
   if( $load > $CPUCRIT )  { $color = "red"; }

   my $summary = "up $up, load = $load, $users users, $procs procs";

   &debug( "cpu - $color, $summary" );
   &status( $SPONGSERVER, $HOST, "cpu", $color, $summary, $message );
}

# Check processes and make sure important ones are around.

sub check_processes { 
   my $color    = "green";
   my( $message, $summary, $check, $ret, $down, @down, $time, $critical );
   
   system( "$PS > $SPONGTMP/PS.$$" );
   foreach $check ( @PROCSWARN ) {
      $ret = `$GREP '$check' $SPONGTMP/PS.$$`;
      if( $ret =~ /^\s*$/ ) { $color = "yellow"; push( @down, $check ); }
      else { $message .= $ret; }
   }
   foreach $check ( @PROCSCRIT ) {
      $ret = `$GREP '$check' $SPONGTMP/PS.$$`;
      if( $ret =~ /^\s*$/ ) { $color = "red"; push( @down, $check ); }
      else { $message .= $ret; }
   }
   unlink "$SPONGTMP/PS.$$";

   $summary = "processes ok"        if $color eq "green";
   $summary = "(@down) not running" if $color ne "green";

   &debug( "procs - $color, $summary" );
   &status( $SPONGSERVER, $HOST, "procs", $color, $summary, $message );
}

# Scan through the log files and flag any messages.

sub check_log_files { 
   my $color    = "green";
   my( $message, $summary, $word, $ret, $hit, @hit, $time, $critical );
   my( %warns, %crits, %tmp );

   # This is way to slow, it scans through the entire log file every time.
   # This wouldn't work very well with a 500M sendmail log file...  I need
   # to do something like remember my position in a log file, and just 
   # scan what is new.

   foreach( keys %LOGWARN, keys %LOGCRIT ) { $tmp{$_}++; }

   foreach $log ( keys %tmp ) {
      my @warnwords = @{$LOGWARN{$log}} if $LOGWARN{$log};
      my @critwords = @{$LOGCRIT{$log}} if $LOGCRIT{$log};
      
      foreach $word ( @warnwords ) {
	 $ret = `$GREP $word $log`;
	 if( $ret !~ /^\s*$/ ) { $color = "yellow"; $warns{$log} .= $ret; }
      }

      foreach $word ( @critwords ) {
	 $ret = `$GREP $word $log`;
	 if( $ret !~ /^\s*$/ ) { $color = "red"; $crits{$log} .= $ret; }
      }
   }
   
   %tmp = ();
   foreach( keys %warns, keys %crits ) { $tmp{$_}++; }
   my @logs = keys %tmp;

   foreach $log ( @logs ) {
      $message .= "$log:\n";
      if( $crits{$log} ) { $message .= $crits{$log} . "\n"; }
      if( $warns{$log} ) { $message .= $warns{$log} . "\n"; }
   }
   
   $summary = "all logs ok" if $color eq "green";
   $summary = "log problems: @logs" if $color ne "green";

   &debug( "logs - $color, $summary" );
   &status( $SPONGSERVER, $HOST, "logs", $color, $summary, $message );
}


# ---------------------------------------------------------------------------
# &status( SERVERADDR, HOST, SERVICE, COLOR, SUMMARY, MESSAGE )
# 
# This function sends information to the Spong server.  It reports the current
# status of a service, and sends along a string of information that might be
# helpful in diagnosing the problem.  This code is modeled after the bb 
# program, but is a little different in that it handles multi-line messages
# and send over the time as an int, rather then a string.
# ---------------------------------------------------------------------------

sub status {
   my( $addr, $host, $cat, $color, $summary, $message ) = @_;
   my( $iaddr, $paddr, $proto, $line, $ip, $ok );

   if( $addr =~ /^\s*((\d+\.){3}\d+)\s*$/ ) {
      $ip = $addr;
   } else {
      my( @addrs ) = (gethostbyname($addr))[4];
      my( $a, $b, $c, $d ) = unpack( 'C4', $addrs[0] );
      $ip = "$a.$b.$c.$d";
   }

   $iaddr = inet_aton( $ip ) || die "no host: $host\n";
   $paddr = sockaddr_in( $SPONGPORT, $iaddr );
   $proto = getprotobyname( 'tcp' );
   
   # Set an alarm so that if we can't connect "immediately" it times out.

   $SIG{'ALRM'} = sub { die };
   alarm(3);

   eval <<'_EOM_';
   socket( SOCK, PF_INET, SOCK_STREAM, $proto ) || die "socket: $!";
   connect( SOCK, $paddr )                      || die "connect: $!";
   select((select(SOCK), $| = 1)[0]);
   print SOCK "status $host $cat $color ", time(), " $summary\n";
   print SOCK "$message\n";
   close( SOCK )                                || die "close: $!";
   $ok = 1;
_EOM_

   alarm(0);
   print STDERR scalar localtime, " can't connect to spong server.\n" if ! $ok;
}


# ---------------------------------------------------------------------------
# Simple debugging function, just accepts a string and if debugging is turned
# on, then the message (along with a timestamp) is sent to stdout.
# ---------------------------------------------------------------------------

sub debug {
   print scalar localtime, " ", $_[0], "\n" if $main::debug;
}


# ---------------------------------------------------------------------------
# Handle some signals...
# ---------------------------------------------------------------------------

sub exit_handler { 
   &debug( "caught QUIT signal, exiting..." );
   unlink "$SPONGTMP/spong-client.pid" if "$SPONGTMP/spong-client.pid";
   exit(0);
}

sub hup_handler {
   &debug( "caught HUP signal, restarting..." );
   unlink "$SPONGTMP/spong-client.pid" if "$SPONGTMP/spong-client.pid";
   if( $debug ) { exec $me, "--debug"; } else { exec $me; }
}

sub pipe_handler {
   wait();
   &debug( "caught $_[0] signal." );
   $SIG{$_[0]} = \&pipe_handler;
}
