#!/usr/local/bin/perl
#
# Spong messaging program.  This script gets called when something that spong
# is monitoring goes red.  This script then runs through a number of tests
# to decided if, who, and how to send a message to someone letting them know
# of the problem.  Currently this script only knows how to send email and 
# send messages to skytel pagers (currently it does it via email as well - but
# I will soon try doing it vi the skytel web server).
#
# History:
# (1) Created (Ed Hill Mar 14, 1997)

use Sys::Hostname;

if( $ARGV[0] eq "--debug" ) { $debug = 1; shift @ARGV; }

$conf_file  = "@@ETCDIR@@/spong.conf";
$hosts_file = "@@ETCDIR@@/spong.hosts";
($HOST)     = gethostbyname(&Sys::Hostname::hostname());
$HOST       =~ tr/A-Z/a-z/;
$ok         = 1;

# Load our configuration variables, including anything specific to the host
# that we are running on.

require $conf_file || die "Can't load $conf_file: $!";
if( -f "$conf_file.$HOST" ) {
   require "$conf_file.$HOST" || die "Can't load $conf_file.$HOST: $!";
} else {
   my $tmp = (split( /\./, $HOST ))[0];
   if( -f "$conf_file.$tmp" ) { # for lazy typist
      require "$conf_file.$tmp" || die "Can't load $conf_file.$tmp: $!";
   }
}
&debug( "spong.conf file(s) loaded" );

# Read in the spong.hosts file.

require $hosts_file || die "Can't load $hosts_file: $!";
&debug( "spong.hosts file loaded" );

# Read the little message history database, so we can prevent message overload

$history_db   = "$SPONGHISTORY/.message-history";
open( HISTORY, $history_db );
while( <HISTORY> ) { chomp; push( @history, $_ ); }
close( HISTORY );
&debug( "message history database loaded" );


# Check usage, and read in what we are supposed to do.

if( $#ARGV == 4 ) {
   ($color, $host, $service, $time, $message) = @ARGV;
} else {
   print STDERR "Error: Incorrect usage!\n";
   print STDERR "Usage: spong-message [--debug] host service time message\n";
   exit(-1);
}

# First find out who we are supposed to contact, and how we are supposed to
# contact them (either email address or pager number).  If there isn't a
# contact or there is not any valid way to reach them, then exit...

($contact, $email, $skytel) = ();
$contact = $HOSTS{$host}->{'contact'};
if( $contact eq "" ) {
   die "Error, no contact information for $host!\n";
} else {
   $email   = $HUMANS{$contact}->{'email'};
   $skytel  = $HUMANS{$contact}->{'skytel'};

   if( $email eq "" && $skytel eq "" ) {
      die "Error, no valid way to contact \"$contact\"!\n";
   }
}
&debug( "contact = $contact, email = \"$email\", skytel = \"$skytel\"" );



# If there is a problem reported, and we are in this machine's down time
# then don't do anything about it.

foreach( @{$main::HOSTS{$host}->{'down'}} ) {
   my( $day, $shour, $smin, $ehour, $emin ) = 
      ( /^(.):(\d+):(\d+)-(\d+):(\d+)$/ );
   my( $nday, $nhour, $nmin ) = (localtime())[6,2,1];

   if( $day eq "*" || $nday eq $day ) {
      my $ntotal = $nhour * 60 + $nmin;
      my $stotal = $shour * 60 + $smin;
      my $etotal = $ehour * 60 + $emin;
      
      if( $ntotal >= $stotal && $ntotal <= $etotal ) {
	 &debug( "problem during downtime, no message will be sent." );
	 $ok = 0;
      }
   }
}
&debug( "not during downtime, continuing" ) if $ok;


# Do the following, in this little loop.  1) Check to see if we have already
# sent this page in the last X minutes.  2) Check to make sure we have sent
# less then X pages to a single contact in the last X minutes.  3) Clean up
# the message history database by removing any items that are over 1 day old.

$mcnt = $ccnt = 0; @okhistory = ();
foreach $item ( @history ) {
   my( $itime, $iok, $icolor, $ihost, $iservice, $icontact ) =
      ( $item =~ /^(\d+)\s+([-+])\s+(\S+)\s+(\S+)\s+(\S+)\s+(.*)$/ );

   if( $time > (time() - 24*60*60) ) { push( @okhistory, $item ); }

   next unless $ok;

   if( $icolor eq $icolor && $ihost eq $host && $iservice eq $service ) {
      if( $itime > (time() - 60*60) && $iok eq "+" ) { $mcnt++; }
   }

   if( $icontact eq $contact && $iok eq "+" ) { 
      if( $itime > (time() - 60*60) ) { $ccnt++; } 
   }
}

if( $ok && $ccnt >= $MESSAGES_PER_HOUR ) { 
   &debug( "$ccnt pages sent to $contact in last hour - no message sent." );
   $ok = 0;
}

if( $ok && $mcnt >= $IDENT_MESSAGES_PER_HOUR ) { 
   &debug( 
     "$mcnt identical pages sent to $contact in last hour - no message sent.");
   $ok = 0;
}

&debug( "checks made, message will be sent" ) if $ok;


# Finally if we can really message someone, then send the message via all
# means that are available (we both page and email if we can).  Of course
# neither of these contact methods is fool-proof, both could fail as a result
# of networking problems or other errors.

if( $ok ) { $okstr = "+"; } else { $okstr = "-"; }

if( $email && $ok ) {
   open( MAIL, "|$SENDMAIL" ) || die "Can't open sendmail: $!";
   print MAIL "To: $email\n";
   print MAIL "Subject: spong - $color $host $service\n\n";
   print MAIL scalar localtime($time), "\n";
   print MAIL "$message\n";
   close( MAIL );
   &debug( "mail message sent to $email" );
}

# The lame way to send a page, I will change this so that it is via their web
# service (at least that way I can verify that the page at least gets to them)

if( $skytel && $ok ) {
   open( MAIL, "|$SENDMAIL" ) || die "Can't open sendmail: $!";
   print MAIL "To: $skytel\@skymail.com\n";
   print MAIL "Subject: spong - $color $host $service\n\n";
   print MAIL scalar localtime($time), "\n";
   print MAIL "$message\n";
   close( MAIL );

   &debug( "skytel page sent to $skytel" );
}


# Write out the message history back to our little one day old message database

open( HISTORY, "> $history_db" );
foreach( sort @okhistory ) { print HISTORY $_, "\n"; }
print HISTORY "$time $okstr $color $host $service $contact\n";
close( HISTORY );
&debug( "message history database updated" );
      
exit(0);


# ---------------------------------------------------------------------------
# Simple debugging function, just accepts a string and if debugging is turned
# on, then the message (along with a timestamp) is sent to stdout.
# ---------------------------------------------------------------------------

sub debug { print scalar localtime, " ", $_[0], "\n" if $main::debug; }
