#!@@PERL@@
#
# Spong network monitoring script.  This runs various tests against IP based
# network services (like nntp, smtp, http, pop, etc...)  If it can not
# connect to a service, then it sends a message to the spong server machine
# about the problems.  This script currently checks the following:
#
#    * ping  (try connecting to it period...)
#    * ftp, pop3, smtp, http, nntp, imap (a suite of tcp based services)
#    * dns (if the Net::DNS module is installed)
#
# I'm not currently using the non-core Net:: tools since that would require an 
# installation of a seperate perl package, but I probably will at some point
#
# History:
# (1) Ported bb-network.sh script to perl. (Ed Hill Feb 27, 1997)

use Sys::Hostname;
use Net::Ping;
use Socket;

# Check to see if they have the Net::DNS module installed, and if they do we
# can then do DNS queries, and see if DNS servers are alive.

eval "require Net::DNS;";
if( ! $@ ) { $dns = 1; } else { $dns = 0; }

srand( time() ^ ($$ + ($$ << 15 )) );

if( $ARGV[0] eq "--debug" )   { $debug = 1;   shift; }
if( $ARGV[0] eq "--restart" ) { $restart = 1; shift; }
if( $ARGV[0] eq "--kill" )    { $kill = 1;    shift; }

$me         = "@@BINDIR@@/spong-network";
$conf_file  = $ARGV[0] || "@@ETCDIR@@/spong.conf";
$hosts_file = "@@ETCDIR@@/spong.hosts";
($HOST)     = gethostbyname(&Sys::Hostname::hostname());
$HOST       =~ tr/A-Z/a-z/;

# Set up some signal handlers to handle our death gracefully, and also listen
# for the HUP signal, and if we se that, we re-exec ourself.

$SIG{'QUIT'} = \&exit_handler;
$SIG{'HUP'}  = \&hup_handler;
$SIG{'PIPE'} = \&pipe_handler;
$SIG{'CHLD'} = \&pipe_handler;

# Load our configuration variables, including anything specific to the host
# that we are running on.

require $conf_file || die "Can't load $conf_file: $!";
if( -f "$conf_file.$HOST" ) {
   require "$conf_file.$HOST" || die "Can't load $conf_file.$HOST: $!";
} else {
   my $tmp = (split( /\./, $HOST ))[0];
   if( -f "$conf_file.$tmp" ) { # for lazy typist
      require "$conf_file.$tmp" || die "Can't load $conf_file.$tmp: $!";
   }
}
&debug( "configuration file(s) loaded" );


# Read in the spong.hosts file.  We are a little nasty here in that we do some
# junk to scan through the file so that we can maintain the order of the hosts
# as they appear in the file.

open( HOSTS, $hosts_file ) || die "Can't load $hosts_file: $!";
while( <HOSTS> ) {
   $evalme .= $_;
   if( /^\s*%HOSTS\s*=\s*\(/ ) { $inhosts = 1; }
   if( $inhosts && /^\s*[\'\"]?([^\s\'\"]+)[\'\"]?\s*\=\>\s*\{/ ) {
      push( @HOSTS_LIST, $1 );
   }
}
close( HOSTS );
eval $evalme || die "Invalid spong.hosts file: $@";

# Fallback, if we didn't read things correctly...
if( sort @HOSTS_LIST != sort keys %HOSTS ) { @HOSTS_LIST = sort keys %HOSTS; }
&debug( "host file loaded" );


# If the user gives us the --restart or --kill flags, then we signal the
# currently running spong-client process, and tell it to either die, or
# re-exec itself (which causes it to re-read it's configuration files.

if( $restart || $kill ) {
   open( PID, "$SPONGTMP/spong-network.pid") || die "Can't find pid file: $!";
   my $pid = <PID>; chomp $pid;
   close PID;
   
   if( $restart ) { 
      &debug( "telling pid $pid to restart" ); kill( 'HUP', $pid ); }
   if( $kill ) { 
      &debug( "telling pid $pid to die" ); kill( 'QUIT', $pid );}

   exit(0);
}

# Write our pid to the spong tmp directory.

open( PID, ">$SPONGTMP/spong-network.pid" ) || 
   print STDERR "Can't create spong-network.pid file: $!\n";
print PID "$$\n"; 
close( PID );

# Do the various network tests for each host.

while( 1 ) {
   foreach $host ( @HOSTS_LIST ) {
      my $checks = $HOSTS{$host}->{'services'};

      &debug( "checking network services ($checks) on $host" );

      &check_ping( $host );

      &check_service( $host, 21, "", "^\s*220", "ftp" )   if $checks =~ /ftp/;
      &check_service( $host, 110, "", "OK", "pop3" )      if $checks =~ /pop/;
      &check_service( $host, 25, "", "^\s*220", "smtp" )  if $checks =~ /smtp/;
      &check_service( $host, 119, "", "^\s*200", "nntp" ) if $checks =~ /nntp/;
      &check_service( $host, 143, "", "OK", "imap" )      if $checks =~ /imap/;

      &check_dns( $host )                                 if $checks =~ /dns/;

      # Http is a little special in that we also check the return code.  No
      # connection and 5xx codes are red, but 4xx return codes are just yellow.
      # Also we go through a list of documents that the have been provided
      # and check each one of them to make sure the web server is behaving
      # correctly

      if( $checks =~ /http/ ) {
	 my( @http_files ) = ( @{$HTTPDOCS{"ALL"}}, @{$HTTPDOCS{$host}} );
	 my( $http_port ) = $HTTPPORT{$host} || $HTTPPORT{"ALL"} || 80;
	 my( $file, $tmessage ) = ( "", "" );
	 my( $color, $summary ) = ( "green", "" );

	 foreach $file ( @http_files ) {
	    my $message = 
	       &check_tcp( $host, $http_port, "HEAD $file HTTP/1.0\r\n\r\n" );

	    if( $message =~ /HTTP\S+\s+(\d\d\d)\s.*$/m ) {
	       my $code = $1;

	       if( $code >= 500 ) {
		  $color = "red"; $summary = "error - $code - $file";
	       } elsif( $code >= 400 ) {
		  if( $color ne "red" ) {
		     $color = "yellow"; $summary = "warning - $code - $file"; }
	       } else { 
		  if( $color ne "red" && $color ne "yellow" ) {
		     $color = "green"; $summary = "ok - $code"; }
	       }
	    } elsif( $message !~ /HTTP/m ) { 
	       $color = "red"; $summary = "no response from http server";
	    } else {
	       if( $color ne "red" ) {
		  $color = "yellow"; $summary = "can't determine status code";}
	    }
	    $tmessage .= "->HEAD $file HTTP/1.0\n$message\n";
	 }

	 &status( $SPONGSERVER, $host, "http", $color, $summary, $tmessage );
      }
   }

   # Sleep about $SPONGSLEEP seconds - add a little bit of randomness so 
   # that things don't get in sync and pound the spong-server.

   my $sleep = int($SPONGSLEEP - (.05 * $SPONGSLEEP) + rand(.1 * $SPONGSLEEP));
   &debug( "sleeping for $sleep seconds" );
   sleep $sleep;
}


# This routine checks connectivity.  It first trys to "ping" the machine via
# a TCP echo using the Net::Ping module, if it can't reach it via that 
# mechanism, then it resorts to the command line ping program.  Using the
# Net::Ping speeds things up quite a bit...

sub check_ping { 
   my( $host ) = @_;
   my( $color, $rt, $summary, $message ) = ( "green", "", "", "" );
   my( @down );

   if( @{$HOSTS{$host}->{'ip_addr'}} ) {
      @hostlist = @{$HOSTS{$host}->{'ip_addr'}};
   } else {
      @hostlist = ( $host );
   }

   foreach $host ( @hostlist ) {
      if( ! pingecho( $host, 15 ) ) {
	 # Try again with a IMCP ping, for those goofy machines that don't run
	 # the echo service...  This is a bit slower...  Make sure your ping
	 # program only sends a finite number of pings...
      
	 my $myping = $PING;
	 my $pingok = 0;
	 $myping =~ s/HOST/$host/g;

	 open( PING, "$myping 2>&1 |") || warn "can't call ping: $!";
	 while( <PING> ) { 
	    $message .= $_; 
	    if( /bytes from/ ) { $pingok = 1; }}
	 close PING;

	 if( ! $pingok ) { 
	    $color = "red"; 
	    $message .= "\n";
	    push( @down, $host );

	    # Provide some traceroute information might be useful in deciding
	    # where the problem is.  This is disabled because it might cause
	    # an unwanted load on your network, and it causes the spong-network
	    # program to spend too much time here...

#	    if( -f $TRACEROUTE ) {
#	       open( PIPE, "$TRACEROUTE $host 2>&1 |" );
#	       while( <PIPE> ) { $message .= $_; }
#	       $message .= "\n";
#	       close( PIPE );
#	    }
	 }
      }
   }

   $summary = "ping failed for " . join( ',', @down ) if $color eq "red";
   $summary = "ping ok"                               if $color eq "green";
   
   &debug( "ping - $host - $color, $summary" );
   &status( $SPONGSERVER, $host, "ping", $color, $summary, $message );
}


# This check will (if the Net::DNS module is available) connect to a DNS server
# and ask that server to resolve it's own name.  If it can do that, then we
# assume it is ok - If it can't then something is wrong.

sub check_dns {
    my( $host ) = @_;
    my( $color, $summary, $message ) = ( "green", "", "" );

    if( ! $dns ) {
       $summary = "can't do DNS lookups, Net::DNS not installed";
       &debug( "dns - $host - $color, $summary" );
       &status( $SPONGSERVER, $host, "dns", "red", $summary,
		"In order to do DNS queries you must install the Net::DNS " .
		"Perl module.\nYou can find the module at your nearest CPAN " .
		"archive or http://www.perl.com/CPAN/\n" );
       return;
    }
       
    my $resolver = new Net::DNS::Resolver;
    my $q = $resolver->search( $host, "A" );

    if( defined $q && defined $q->answer && defined (($q->answer)[0]) ) {
       $color = "green";
       $summary = "dns ok";
    } else {
       $color = "red";
       $summary = "can't resolve $host";
       $message = "can't resolve $host\n";
    }

    &debug( "dns - $host - $color, $summary" );
    &status( $SPONGSERVER, $host, "dns", $color, $summary, $message );
}


# A generic tcp port checking routine.  You give this function a hostname, a
# port, a message to send (can be ""), a return regular expression string to 
# check for, and the name of the service.  This will go out connect to that
# port and check to make sure you get back expected results.

sub check_service {
   my( $host, $port, $send, $check, $service ) = @_;
   my( $color, $summary ) = ( "green", "" );

   my $start   = time();
   my $message = &check_tcp( $host, $port, $send );
   my $diff    = time() - $start;

   if( $message !~ /$check/ ) { $color = "red"; }

   $summary = "$service down"                            if $color eq "red";
   $summary = "$service ok - $diff second response time" if $color eq "green";
      
   &debug( "$service - $host - $color, $summary" );
   &status( $SPONGSERVER, $host, $service, $color, $summary, $message );
}


# ---------------------------------------------------------------------------
# &status( SERVERADDR, HOST, SERVICE, COLOR, SUMMARY, MESSAGE )
# 
# This function sends information to the Spong server.  It reports the current
# status of a service, and sends along a string of information that might be
# helpful in diagnosing the problem.  This code is modeled after the bb 
# program, but is a little different in that it handles multi-line messages
# and send over the time as an int, rather then a string.
# ---------------------------------------------------------------------------

sub status {
   my( $addr, $host, $cat, $color, $summary, $message ) = @_;
   my( $iaddr, $paddr, $proto, $line, $ip, $ok );

   if( $addr =~ /^\s*((\d+\.){3}\d+)\s*$/ ) {
      $ip = $addr;
   } else {
      my( @addrs ) = (gethostbyname($addr))[4];
      my( $a, $b, $c, $d ) = unpack( 'C4', $addrs[0] );
      $ip = "$a.$b.$c.$d";
   }

   $iaddr = inet_aton( $ip ) || die "no host: $host\n";
   $paddr = sockaddr_in( $SPONGPORT, $iaddr );
   $proto = getprotobyname( 'tcp' );
   
   # Set an alarm so that if we can't connect "immediately" it times out.

   $SIG{'ALRM'} = sub { die };
   alarm(3);

   eval <<'_EOM_';
   socket( SOCK, PF_INET, SOCK_STREAM, $proto ) || die "socket: $!";
   connect( SOCK, $paddr )                      || die "connect: $!";
   select((select(SOCK), $| = 1)[0]);
   print SOCK "status $host $cat $color ", time(), " $summary\n";
   print SOCK "$message\n";
   close( SOCK )                                || die "close: $!";
   $ok = 1;
_EOM_

   alarm(0);
   print STDERR scalar localtime, " can't connect to spong server.\n" if ! $ok;
}

# ---------------------------------------------------------------------------
# &check_tcp( HOST, PORT, DATA )
#
# This function will make a connection to a port at a given port, and send a
# message, it will then return what it gets back to the caller of this
# function.
# ---------------------------------------------------------------------------

sub check_tcp {
   my( $addr, $port, $data ) = @_;
   my( $iaddr, $paddr, $proto, $line, $ip );

   if( $addr =~ /^\s*((\d+\.){3}\d+)\s*$/ ) {
      $ip = $addr;
   } else {
      my( @addrs ) = (gethostbyname($addr))[4];
      my( $a, $b, $c, $d ) = unpack( 'C4', $addrs[0] );
      $ip = "$a.$b.$c.$d";
   }

   $iaddr = inet_aton( $ip )                    || return -1;
   $paddr = sockaddr_in( $port, $iaddr );
   $proto = getprotobyname( 'tcp' );
 
   # Set an alarm so that if we can't connect "immediately" it times out.
   # Poor man's exception handling in perl...

   $SIG{'ALRM'} = sub { die };
   alarm(30);

   eval <<'_EOM_';
   socket( SOCK, PF_INET, SOCK_STREAM, $proto ) || return -2;
   connect( SOCK, $paddr )                      || return -3;
   select((select(SOCK), $| = 1)[0]);
   print SOCK "$data";
   recv( SOCK, $line, 256, 0 );  # just grab a chunk from the service.
   close( SOCK )                                || return -4;
_EOM_
   
   alarm(0);
   return $line;
}


# ---------------------------------------------------------------------------
# Simple debugging function, just accepts a string and if debugging is turned
# on, then the message (along with a timestamp) is sent to stdout.
# ---------------------------------------------------------------------------

sub debug {
   print scalar localtime, " ", $_[0], "\n" if $main::debug;
}

# ---------------------------------------------------------------------------
# Handle some signals...
# ---------------------------------------------------------------------------

sub exit_handler { 
   &debug( "caught QUIT signal, exiting..." );
   unlink "$SPONGTMP/spong-network.pid" if "$SPONGTMP/spong-network.pid";
   exit(0);
}

sub hup_handler {
   &debug( "caught HUP signal, restarting..." );
   unlink "$SPONGTMP/spong-network.pid" if "$SPONGTMP/spong-network.pid";
   if( $debug ) { exec $me, "--debug"; } else { exec $me; }
}

sub pipe_handler {
   wait();
   &debug( "caught $_[0] signal." );
   $SIG{$_[0]} = \&pipe_handler;
}
