#!@@PERL@@
#
# Spong server daemon.  This listens at port 1984, accepts input from client
# connections and logs it to a file.  It doesn't respond with any 
# acknowledgements.  It also maintains a history of what problems occur.
#
# History:
# (1) Ported bbd.c script to perl. (Ed Hill Feb 28, 1997)

use Sys::Hostname;
use Socket;

if( $ARGV[0] eq "--debug" )   { $debug = 1;   shift; }
if( $ARGV[0] eq "--restart" ) { $restart = 1; shift; }
if( $ARGV[0] eq "--kill" )    { $kill = 1;    shift; }

$me         = "@@BINDIR@@/spong-server";
$smessage   = "@@BINDIR@@/spong-message";
$conf_file  = $ARGV[0] || "@@ETCDIR@@/spong.conf";
$hosts_file = "@@ETCDIR@@/spong.hosts";
($HOST)     = gethostbyname(&Sys::Hostname::hostname());
$HOST       =~ tr/A-Z/a-z/;

# Set up some signal handlers to handle our death gracefully, and also listen
# for the HUP signal, and if we se that, we re-exec ourself.

$SIG{'QUIT'} = \&exit_handler;
$SIG{'HUP'}  = \&hup_handler;

# Load our configuration variables, including anything specific to the host
# that we are running on.

require $conf_file || die "Can't load $conf_file: $!";
if( -f "$conf_file.$HOST" ) {
   require "$conf_file.$HOST" || die "Can't load $conf_file.$HOST: $!";
} else {
   my $tmp = (split( /\./, $HOST ))[0];
   if( -f "$conf_file.$tmp" ) { # for lazy typist
      require "$conf_file.$tmp" || die "Can't load $conf_file.$tmp: $!";
   }
}
&debug( "configuration file(s) loaded" );

# Read in the spong.hosts file.

require $hosts_file || die "Can't load $hosts_file: $!";
&debug( "host file loaded" );

# If the user gives us the --restart or --kill flags, then we signal the
# currently running spong-client process, and tell it to either die, or
# re-exec itself (which causes it to re-read it's configuration files.

if( $restart || $kill ) {
   open( PID, "$SPONGTMP/spong-server.pid") || die "Can't find pid file: $!";
   my $pid = <PID>; chomp $pid;
   close PID;
   
   if( $restart ) { 
      &debug( "telling pid $pid to restart" ); kill( 'HUP', $pid ); }
   if( $kill ) { 
      &debug( "telling pid $pid to die" ); kill( 'QUIT', $pid );}

   exit(0);
}

# Write our pid to the spong tmp directory.

open( PID, ">$SPONGTMP/spong-server.pid" ) || 
   print STDERR "Can't create spong-server.pid file: $!\n";
print PID "$$\n"; 
close( PID );


# Set up the socket to listen to

my $proto = getprotobyname( 'tcp' );
my $p     = pack( "l", 1 );
socket( SERVER, PF_INET, SOCK_STREAM, $proto )        || die "socket: $!";
setsockopt( SERVER, SOL_SOCKET, SO_REUSEADDR, $p )    || die "setsockopt: $!";
bind( SERVER, sockaddr_in( $SPONGPORT, INADDR_ANY ) ) || die "bind: $!";
listen( SERVER, SOMAXCONN )                           || die "listen: $!";

sub reaper { 
   $waitedpid = wait(); 
   $SIG{'CHLD'} = \&reaper; 
   &debug( "reaping $waitedpid with exit $?" );
};
$SIG{'CHLD'} = \&reaper;

&debug( "server socket setup, listening for connections" );

# Start listening at our socket, and hand off any connections to a child
# process to deal with.

for( $waitedpid = 0; 
     ($paddr = accept(CLIENT,SERVER)) || $waitedpid;
     $waitedpid = 0, close CLIENT ) {

   if( $waitedpid && ! $paddr ) {
#      print "skipping ($waitedpid)...\n";
      next;
   }

   # Fork off a child to handle the request, the server just goes right back
   # to listening to our master socket.

   if( ! defined( $pid = fork() ) ) {
      print STDERR "Warning: can not fork: $!"; next;
   } elsif( $pid ) {
      if( ($waitedpid = wait()) > 0 ) {
	 &debug( "Hmmm.... reaping $waitedpid with exit $?" ); }
      next;
   }

   # I'm the child, read the input from the CLIENT

   my( $port, $iaddr ) = sockaddr_in( $paddr );
   my $name            = gethostbyaddr( $iaddr, AF_INET );
   my $ip              = inet_ntoa( $iaddr );

   &debug( "[$$] connection from $name ($ip)" );

   # I really should do some kind of check here to make sure the message
   # is coming from a valid client.  A simple check would be to make sure 
   # that the connection is coming from a host in the SPONG-hosts file, but
   # that is really not good enough...

   my $header = <CLIENT>; chomp $header;
   my( $message, $cnt, $line ) = ( "", "", "" );
   while( defined( $line = <CLIENT> ) ) { 
      $message .= $line; 
      $cnt += length($line);
      if( $cnt > 100000 ) { last; }
   }
   close( CLIENT );

   # Do some checking on the message.  If it appears bogus, then just
   # log a message, and close the connection.

   if( $header =~ /^(\w+) (\S+) (\w+) (\w+) (\d+) (.*)$/ ) {
      my( $cmd, $host, $type, $color, $time, $sum ) = ($1, $2, $3, $4, $5, $6);
      $host =~ tr/a-zA-Z0-9_\-\.//cd;
      $type =~ tr/a-zA-Z0-9_\-\.//cd;

      if( $color ne "red" && $color ne "yellow" && $color ne "green" ) {
	 print STDERR scalar localtime, " invalid message: '$header'\n";
	 exit(0);
      }

      # If we are supposed to act in some way on the message (like sending a
      # page), then do that, otherwise just call the logging function that
      # saves the information sent to us locally.

      if( $cmd eq "status" ) {
	 if( $SEND_MESSAGE eq "RED" ) {
	    if( $color eq "red" ) {
	       if( -f $smessage ) {
		  &debug( "[$$] color = red, messaging a human" );
		  system $smessage, $color, $host, $type, $time, $sum;
	       } else {
		  print STDERR scalar localtime, 
	          " could not send page, $smessage not found\n";
	       }
	    }
	 } elsif( $SEND_MESSAGE eq "CHANGE" ) {
	    # If the saved state of this service is different then it is
	    # now, then send a message.

	    if( ! -f "$SPONGSTATUS/$host-$type-$color" ) {
	       if( -f $smessage ) {
		  &debug( "[$$] change in state, messaging a human" );
		  system $smessage, $color, $host, $type, $time, $sum;
	       } else {
		  print STDERR scalar localtime, 
		  " could not send page, $smessage not found\n";
	       }
	    }
	 }
	    
	 &save_message( $host, $type, $color, $time, $sum, $message );
      } else {
	 print STDERR scalar localtime, " invalid message: '$header'\n";
      }
   }

   exit(0);
}


# ---------------------------------------------------------------------------
# Take the incoming message, check to see if it is different from the last
# message, and if so - update the history, then save the information locally
# in the status directory.
# ---------------------------------------------------------------------------

sub save_message {
   my( $host, $type, $color, $time, $sum, $message ) = @_;

   if( -d $SPONGSTATUS ) {
      # If the status of this service on this host has changed, then record
      # a record to a history file.

      if( ! -f "$SPONGSTATUS/$host-$type-$color" ) {
	 &debug( "[$$] updating history for $host/$type/$color" );

	 open( HISTORY, ">> $SPONGHISTORY/$host-$type" ) ||
	    print STDERR scalar localtime, 
 	       " can't update $SPONGHISTORY/$host-$type: $!\n";
         select((select(HISTORY), $| = 1)[0]);
	 print HISTORY "$time $color", " "x(7-length($color)), "$sum\n";
	 close( HISTORY );
	 chmod 0644, "$SPONGHISTORY/$host-$type";
      }
      
      foreach( "red", "yellow", "green" ) {
	 if( -f "$SPONGSTATUS/$host-$type-$_" ) {
	    unlink "$SPONGSTATUS/$host-$type-$_";
	 }
      }
  
      $message =~ s/(\r)//g;

      &debug( "[$$] updating status for $host/$type/$color" );
      open( STATUS, "> $SPONGSTATUS/$host-$type-$color" ) ||
	 print STDERR scalar localtime, 
	    " can't create $SPONGSTATUS/$host-$type-$color: $!\n";
      select((select(STATUS), $| = 1)[0]);
      print STATUS "$time $sum\n$message";
      close( STATUS );
      chmod 0644, "$SPONGSTATUS/$host-$type-$color";
   } else {
      die "Invalid status directory: $SPONGSTATUS";
   }
}


# ---------------------------------------------------------------------------
# Simple debugging function, just accepts a string and if debugging is turned
# on, then the message (along with a timestamp) is sent to stdout.
# ---------------------------------------------------------------------------

sub debug {
   print scalar localtime, " ", $_[0], "\n" if $main::debug;
}

# ---------------------------------------------------------------------------
# Handle some signals...
# ---------------------------------------------------------------------------

sub exit_handler { 
   &debug( "caught QUIT signal, exiting..." );
   unlink "$SPONGTMP/spong-server.pid" if "$SPONGTMP/spong-server.pid";
   exit(0);
}

sub hup_handler {
   &debug( "caught HUP signal, restarting..." );
   unlink "$SPONGTMP/spong-server.pid" if "$SPONGTMP/spong-server.pid";
   if( $debug ) { exec $me, "--debug"; } else { exec $me; }
}
